from typing import Awaitable, Callable, Dict
from rekuest_next.messages import Assign
from fluss_next.api.schema import (
    RekuestFilterNode,
    RekuestMapNode,
    ReactiveNode,
    BaseGraphNodeBase,
    MapStrategy,
    ReactiveImplementation,
    NodeKind,
)
import asyncio
from reaktion_next.atoms.arkitekt import (
    ArkitektMapAtom,
    ArkitektMergeMapAtom,
    ArkitektAsCompletedAtom,
    ArkitektOrderedAtom,
)
from reaktion_next.atoms.arkitekt_filter import ArkitektFilterAtom
from reaktion_next.atoms.transformation.chunk import ChunkAtom
from reaktion_next.atoms.transformation.buffer_complete import BufferCompleteAtom
from reaktion_next.atoms.transformation.split import SplitAtom
from reaktion_next.atoms.combination.zip import ZipAtom
from reaktion_next.atoms.transformation.filter import FilterAtom
from reaktion_next.atoms.combination.withlatest import WithLatestAtom
from reaktion_next.atoms.combination.gate import GateAtom
from reaktion_next.atoms.filter.all import AllAtom
from rekuest_next.postmans.contract import RPCContract
from .base import Atom
from .transport import AtomTransport
from rekuest_next.messages import Assign
from typing import Any, Optional
from reaktion_next.atoms.operations.math import MathAtom, operation_map


def atomify(
    node: BaseGraphNodeBase,
    transport: AtomTransport,
    contract: Optional[RPCContract],
    globals: Dict[str, Any],
    assignment: Assign,
    alog: Callable[[Assign, str, str], Awaitable[None]] = None,
) -> Atom:
    if isinstance(node, RekuestMapNode):
        if node.node_kind == NodeKind.FUNCTION:
            if node.map_strategy == MapStrategy.MAP:
                return ArkitektMapAtom(
                    node=node,
                    contract=contract,
                    transport=transport,
                    assignment=assignment,
                    globals=globals,
                    alog=alog,
                )
            if node.map_strategy == MapStrategy.AS_COMPLETED:
                return ArkitektAsCompletedAtom(
                    node=node,
                    contract=contract,
                    transport=transport,
                    assignment=assignment,
                    globals=globals,
                    alog=alog,
                )
            if node.map_strategy == MapStrategy.ORDERED:
                return ArkitektAsCompletedAtom(
                    node=node,
                    contract=contract,
                    transport=transport,
                    assignment=assignment,
                    globals=globals,
                    alog=alog,
                )

            raise NotImplementedError(
                f"Map strategy {node.map_strategy} is not implemented"
            )
        if node.node_kind == NodeKind.GENERATOR:
            return ArkitektMergeMapAtom(
                node=node,
                contract=contract,
                transport=transport,
                assignment=assignment,
                globals=globals,
                alog=alog,
            )

        raise NotImplementedError(f"Node kind {node.kind} is not implemented")
    if isinstance(node, RekuestFilterNode):
        if node.node_kind == NodeKind.FUNCTION:
            if node.map_strategy == MapStrategy.MAP:
                return ArkitektFilterAtom(
                    node=node,
                    contract=contract,
                    transport=transport,
                    assignment=assignment,
                    globals=globals,
                    alog=alog,
                )
        if node.node_kind == NodeKind.GENERATOR:
            raise NotImplementedError("Generator cannot be used as a filter")

    if isinstance(node, ReactiveNode):
        if node.implementation == ReactiveImplementation.ZIP:
            return ZipAtom(
                node=node,
                transport=transport,
                assignment=assignment,
                globals=globals,
                alog=alog,
            )
        if node.implementation == ReactiveImplementation.FILTER:
            return FilterAtom(
                node=node,
                transport=transport,
                assignment=assignment,
                globals=globals,
                alog=alog,
            )
        if node.implementation == ReactiveImplementation.CHUNK:
            return ChunkAtom(
                node=node,
                transport=transport,
                assignment=assignment,
                globals=globals,
                alog=alog,
            )
        if node.implementation == ReactiveImplementation.GATE:
            return GateAtom(
                node=node,
                transport=transport,
                assignment=assignment,
                globals=globals,
                alog=alog,
            )
        if node.implementation == ReactiveImplementation.BUFFER_COMPLETE:
            return BufferCompleteAtom(
                node=node,
                transport=transport,
                assignment=assignment,
                globals=globals,
                alog=alog,
            )
        if node.implementation == ReactiveImplementation.WITHLATEST:
            return WithLatestAtom(
                node=node,
                transport=transport,
                assignment=assignment,
                globals=globals,
                alog=alog,
            )
        if node.implementation == ReactiveImplementation.COMBINELATEST:
            return WithLatestAtom(
                node=node,
                transport=transport,
                assignment=assignment,
                globals=globals,
                alog=alog,
            )
        if node.implementation == ReactiveImplementation.SPLIT:
            return SplitAtom(
                node=node,
                transport=transport,
                assignment=assignment,
                globals=globals,
                alog=alog,
            )
        if node.implementation == ReactiveImplementation.ALL:
            return AllAtom(
                node=node,
                transport=transport,
                assignment=assignment,
                globals=globals,
                alog=alog,
            )
        if node.implementation in operation_map:
            return MathAtom(
                node=node,
                transport=transport,
                assignment=assignment,
                globals=globals,
                alog=alog,
            )

    raise NotImplementedError(f"Atom for {node} {type(node)} is not implemented")



import asyncio
from typing import List
from reaktion_next.atoms.combination.base import CombinationAtom
from reaktion_next.events import EventType, OutEvent
import logging

logger = logging.getLogger(__name__)


class SplitAtom(CombinationAtom):
    complete: List[bool] = [False, False]

    async def run(self):
        try:
            while True:
                event = await self.get()

                if event.type == EventType.ERROR:
                    for index, stream in enumerate(self.node.outstream):
                        await self.transport.put(
                            OutEvent(
                                handle=f"return_{index}",
                                type=EventType.ERROR,
                                exception=event.exception,
                                source=self.node.id,
                                caused_by=[event.current_t],
                            )
                        )
                    break

                if event.type == EventType.NEXT:
                    for index, stream in enumerate(self.node.outstream):
                        if event.value[index] is not None:
                            await self.transport.put(
                                OutEvent(
                                    handle=f"return_{index}",
                                    type=EventType.NEXT,
                                    value=(event.value[index],),
                                    source=self.node.id,
                                    caused_by=[event.current_t],
                                )
                            )

                if event.type == EventType.COMPLETE:
                    for index, stream in enumerate(self.node.outstream):
                        await self.transport.put(
                            OutEvent(
                                handle=f"return_{index}",
                                type=EventType.COMPLETE,
                                value=None,
                                source=self.node.id,
                                caused_by=[event.current_t],
                            )
                        )
                    break

        except asyncio.CancelledError as e:
            logger.warning(f"Atom {self.node} is getting cancelled")
            raise e

        except Exception as e:
            logger.exception(f"Atom {self.node} excepted")
            raise e



import asyncio
from typing import List
from reaktion_next.atoms.transformation.base import TransformationAtom
from reaktion_next.events import EventType, OutEvent, InEvent
import logging
from pydantic import Field
from functools import reduce

logger = logging.getLogger(__name__)


class AllAtom(TransformationAtom):
    def assert_values(self, values, check_list_length=True):
        if not isinstance(values, list):
            raise ValueError("AllAtom expects a list of values")

        if not all([v is not None for v in values]):
            raise ValueError("Atom expects all values to be not None")

        if check_list_length:
            if not all([len(v) != 0 for v in values if isinstance(v, List)]):
                raise ValueError("Atom expects all lists to be not empty")

    async def run(self):
        list_length = self.set_values.get("list_length", False)
        try:
            while True:
                event = await self.get()

                if event.type == EventType.ERROR:
                    await self.transport.put(
                        OutEvent(
                            handle="return_0",
                            type=EventType.ERROR,
                            exception=event.exception,
                            source=self.node.id,
                            caused_by=[event.current_t],
                        )
                    )
                    break

                if event.type == EventType.COMPLETE:
                    await self.transport.put(
                        OutEvent(
                            handle="return_0",
                            type=EventType.COMPLETE,
                            source=self.node.id,
                            caused_by=[event.current_t],
                        )
                    )
                    break

                if event.type == EventType.NEXT:
                    try:
                        self.assert_values(event.value, check_list_length=list_length)
                        await self.transport.put(
                            OutEvent(
                                handle="return_0",
                                type=EventType.NEXT,
                                value=event.value,
                                source=self.node.id,
                                caused_by=[event.current_t],
                            )
                        )
                    except ValueError as e:
                        logger.exception(f"Atom {self.node} filtered out an event")

                        await self.transport.put(
                            OutEvent(
                                handle="return_1",
                                type=EventType.NEXT,
                                value=event.value,
                                source=self.node.id,
                                caused_by=[event.current_t],
                            )
                        )

        except asyncio.CancelledError as e:
            logger.warning(f"Atom {self.node} is getting cancelled")
            raise e

        except Exception as e:
            logger.exception(f"Atom {self.node} excepted")
            raise e



from fluss_next.api.schema import GraphNodeBase
from reaktion_next.events import InEvent


def index_for_handle(handle: str) -> int:
    return int(handle.split("_")[1])


def node_to_reference(node: GraphNodeBase, event: InEvent) -> str:
    return f"{node.id}_{event.current_t}"



from typing import List, Optional
from rekuest_next.api.schema import DefinitionInput, PortInput, NodeKind
from fluss_next.api.schema import (
    Graph,
    RekuestNodeBase,
    ReactiveNode,
    ReactiveImplementation,
    ArgNode,
    ReturnNode,
    GlobalArg,
    Flow,
)
from .events import OutEvent, InEvent
import pydantic
from .errors import FlowLogicError


def connected_events(graph: Graph, event: OutEvent, t: int) -> List[InEvent]:
    events = []

    for edge in graph.edges:
        if edge.source == event.source and edge.source_handle == event.handle:
            try:
                events.append(
                    InEvent(
                        target=edge.target,
                        handle=edge.target_handle,
                        type=event.type,
                        value=event.value,
                        exception=event.exception,
                        current_t=t,
                    )
                )
            except pydantic.ValidationError as e:
                raise FlowLogicError(f"Invalid event for {edge} : {event}") from e

    return events


def infer_kind_from_graph(graph: Graph) -> NodeKind:
    kind = NodeKind.FUNCTION

    for node in graph.nodes:
        if isinstance(node, RekuestNodeBase):
            if node.kind == NodeKind.GENERATOR:
                kind = NodeKind.GENERATOR
                break
        if isinstance(node, ReactiveNode):
            if node.implementation == ReactiveImplementation.CHUNK:
                kind = NodeKind.GENERATOR
                break

    return kind


def convert_flow_to_definition(
    flow: Flow,
    name: str = None,
    description: str = None,
    kind: Optional[NodeKind] = None,
) -> DefinitionInput:
    # assert localnodes are in the definitionregistry

    argNode = [x for x in flow.graph.nodes if isinstance(x, ArgNode)][0]
    returnNode = [x for x in flow.graph.nodes if isinstance(x, ReturnNode)][0]

    args = [PortInput(**x.dict(by_alias=True)) for x in argNode.outs[0]]
    returns = [PortInput(**x.dict(by_alias=True)) for x in returnNode.ins[0]]

    globals = [
        PortInput(**glob.port.dict(by_alias=True)) for glob in flow.graph.globals
    ]

    return DefinitionInput(
        name=name or flow.title,
        kind=kind or infer_kind_from_graph(flow.graph),
        args=args + globals,
        returns=returns,
        portGroups=[],
        description=description,
        interfaces=[
            "workflow",
            f"diagram:{flow.workspace.id}",
            f"flow:{flow.id}",
        ],
    )
